from enum import Enum
from typing import Literal, Optional, TypedDict, Union

from outspeed.plugins.openai_realtime.events import ServerEvent


class AudioFormatType(str, Enum):
    PCM16 = "pcm16"
    G711_ULAW = "g711-ulaw"
    G711_ALAW = "g711-alaw"


class FunctionCallType(str, Enum):
    FUNCTION = "function"


class ItemStatusType(str, Enum):
    IN_PROGRESS = "in_progress"
    COMPLETED = "completed"
    INCOMPLETE = "incomplete"


class RoleType(str, Enum):
    SYSTEM = "system"
    USER = "user"
    ASSISTANT = "assistant"


class SessionVoiceType(str, Enum):
    ALLY = "alloy"
    SHIMMER = "shimmer"
    ECHO = "echo"


class ToolChoiceType(str, Enum):
    AUTO = "auto"
    NONE = "none"
    REQUIRED = "required"


class InputAudioTranscription(TypedDict):
    model: str


class ServerVad(TypedDict):
    type: Literal["server_vad"]
    threshold: Optional[float]
    prefix_padding_ms: Optional[int]
    silence_duration_ms: Optional[int]


class FunctionTool(TypedDict):
    type: Literal["function"]
    name: str
    description: Optional[str]
    parameters: dict


class Session(TypedDict):
    id: str
    object: Literal[ServerEvent.SESSION_CREATED]
    expires_at: int
    model: str
    modalities: list[Literal["text", "audio"]]
    instructions: str
    voice: SessionVoiceType
    input_audio_format: AudioFormatType
    output_audio_format: AudioFormatType
    input_audio_transcription: Optional[InputAudioTranscription]
    turn_detection: Optional[ServerVad]
    tools: list[FunctionTool]
    tool_choice: ToolChoiceType
    temperature: float
    max_response_output_tokens: Optional[int]


class TextContent(TypedDict):
    type: Literal["text"]
    text: str


class InputTextContent(TypedDict):
    type: Literal["input_text"]
    text: str


class AudioContent(TypedDict):
    type: Literal["audio"]
    audio: str  # b64


class InputAudioContent(TypedDict):
    type: Literal["input_audio"]
    audio: str  # b64


class FunctionTool(TypedDict):
    type: Literal["function"]
    name: str
    description: Optional[str]
    parameters: dict


class SystemItem(TypedDict):
    id: str
    object: Literal["realtime.item"]
    type: Literal["message"]
    role: Literal["system"]
    content: list[InputTextContent]


class UserItem(TypedDict):
    id: str
    object: Literal["realtime.item"]
    type: Literal["message"]
    role: Literal["user"]
    content: list[Union[InputTextContent, InputAudioContent]]


class AssistantItem(TypedDict):
    id: str
    object: Literal["realtime.item"]
    type: Literal["message"]
    role: Literal["assistant"]
    content: list[Union[TextContent, AudioContent]]


class FunctionCallItem(TypedDict):
    id: str
    object: Literal["realtime.item"]
    type: Literal["function_call"]
    call_id: str
    name: str
    arguments: str


class FunctionCallOutputItem(TypedDict):
    id: str
    object: Literal["realtime.item"]
    type: Literal["function_call_output"]
    call_id: str
    output: str


class SessionCreated(TypedDict):
    event_id: str
    type: Literal[ServerEvent.SESSION_CREATED]
    session: Session


class SessionUpdated(TypedDict):
    event_id: str
    type: Literal[ServerEvent.SESSION_UPDATED]
    session: Session


ConversationItem = Union[SystemItem, UserItem, FunctionCallItem, FunctionCallOutputItem]


class ConversationItemCreated(TypedDict):
    event_id: str
    type: Literal[ServerEvent.CONVERSATION_ITEM_CREATED]
    item: ConversationItem


class ResponseAudioDelta(TypedDict):
    event_id: str
    type: Literal[ServerEvent.RESPONSE_AUDIO_DELTA]
    response_id: str
    output_index: int
    content_index: int
    delta: str  # b64


class InputAudioBufferSpeechStarted(TypedDict):
    event_id: str
    type: Literal[ServerEvent.INPUT_AUDIO_BUFFER_SPEECH_STARTED]
    item_id: str
    audio_start_ms: int


class ResponseAudioTranscriptDelta(TypedDict):
    event_id: str
    type: Literal[ServerEvent.RESPONSE_AUDIO_TRANSCRIPT_DELTA]
    response_id: str
    output_index: int
    content_index: int
    delta: str
    item_id: str


class CancelledStatusDetails(TypedDict):
    type: Literal["cancelled"]
    reason: Literal["turn_detected", "client_cancelled"]


class IncompleteStatusDetails(TypedDict):
    type: Literal["incomplete"]
    reason: Literal["max_output_tokens", "content_filter"]


class Error(TypedDict):
    code: str
    message: str


class FailedStatusDetails(TypedDict):
    type: Literal["failed"]
    error: Optional[Error]


ResponseStatusDetails = Union[CancelledStatusDetails, IncompleteStatusDetails, FailedStatusDetails]


class Usage(TypedDict):
    total_tokens: int
    input_tokens: int
    output_tokens: int


class Response(TypedDict):
    id: str
    object: Literal["realtime.response"]
    status: Literal["in_progress", "completed", "incomplete", "cancelled", "failed"]
    status_details: Optional[ResponseStatusDetails]
    output: list[ConversationItem]
    usage: Optional[Usage]


class ResponseTextDeltaAdded(TypedDict):
    event_id: str
    type: Literal[ServerEvent.RESPONSE_TEXT_DELTA]
    response_id: str
    output_index: int
    content_index: int
    delta: str
    item_id: str


class Conversation(TypedDict):
    id: str
    object: Literal["realtime.conversation"]


class ConversationCreated(TypedDict):
    event_id: str
    type: Literal[ServerEvent.CONVERSATION_CREATED]
    conversation: Conversation


class ConversationItemInputAudioTranscriptionCompleted(TypedDict):
    event_id: str
    type: Literal[ServerEvent.CONVERSATION_ITEM_INPUT_AUDIO_TRANSCRIPTION_COMPLETED]
    item_id: str
    content_index: int
    transcript: str


class ContentPart(TypedDict):
    type: Literal["text", "audio"]
    audio: Optional[str]  # b64
    transcript: Optional[str]


class ResponseContentPartAdded(TypedDict):
    event_id: str
    type: Literal[ServerEvent.RESPONSE_CONTENT_PART_ADDED]
    item_id: str
    response_id: str
    output_index: int
    content_index: int
    part: ContentPart


class ResponseDone(TypedDict):
    event_id: str
    type: Literal[ServerEvent.RESPONSE_DONE]
    response: Response


class ResponseFunctionCallArgumentsDone(TypedDict):
    event_id: str
    type: Literal[ServerEvent.RESPONSE_FUNCTION_CALL_ARGUMENTS_DONE]
    response_id: str
    output_index: int
    arguments: str
    name: str
    item_id: str
    call_id: str



import asyncio
from typing import Type

import openai
from pydantic import BaseModel


class Tool:
    def __init__(
        self,
        name: str,
        description: str,
        parameters_type: Type[BaseModel],
        response_type: Type[BaseModel],
        should_respond_after_tool_call: bool = False,
    ):
        self.name = name
        self.description = description

        self.parameters_type = parameters_type
        self.response_type = response_type
        self.should_respond_after_tool_call = should_respond_after_tool_call

    async def run(self, input_parameters: Type[BaseModel]):
        raise NotImplementedError("Tool run method must be implemented by the subclass")

    def to_openai_tool_json(self):
        json = openai.pydantic_function_tool(
            model=self.parameters_type,
            name=self.name,
            description=self.description,
        )
        return json

    def to_openai_tool_response_json(self, response: BaseModel):
        json = openai.pydantic_function_tool(
            model=self.response_type,
            name=self.name,
            description=self.description,
        )
        return json

    async def _run_tool(self, function_json: dict):
        if function_json["function"]["name"] != self.name:
            raise ValueError(f"Tool name mismatch: {function_json['name']} != {self.name}")

        input_parameters = self.parameters_type.model_validate(function_json["function"]["arguments"])
        response = await self.run(input_parameters)

        if not isinstance(response, self.response_type):
            raise ValueError(f"Tool response type mismatch: {type(response)} != {self.response_type}")

        json = {"tool_call_id": function_json["id"], "role": "tool", "content": response.model_dump_json()}

        return json



import asyncio
from typing import List

from outspeed.streams import AudioStream, ByteStream, Stream, TextStream, VideoStream


def combine_latest(input_queues: List[Stream]):
    output_queues = []
    for q in input_queues:
        if isinstance(q, AudioStream):
            output_queues.append(AudioStream())
        elif isinstance(q, VideoStream):
            output_queues.append(VideoStream())
        elif isinstance(q, TextStream):
            output_queues.append(TextStream())
        elif isinstance(q, ByteStream):
            output_queues.append(ByteStream())

    async def run():
        while True:
            if any([q.empty() for q in input_queues]):
                await asyncio.sleep(0.2)
                continue

            # Dequeue from input and enqueue to output simultaneously
            for in_q, out_q in zip(input_queues, output_queues):
                item = in_q.get_nowait()
                out_q.put_nowait(item)

    asyncio.create_task(run())
    return output_queues



from enum import Enum


class ServerEvent(str, Enum):
    ERROR = "error"
    SESSION_CREATED = "session.created"
    SESSION_UPDATED = "session.updated"
    CONVERSATION_CREATED = "conversation.created"
    INPUT_AUDIO_BUFFER_COMMITTED = "input_audio_buffer.committed"
    INPUT_AUDIO_BUFFER_CLEARED = "input_audio_buffer.cleared"
    INPUT_AUDIO_BUFFER_SPEECH_STARTED = "input_audio_buffer.speech_started"
    INPUT_AUDIO_BUFFER_SPEECH_STOPPED = "input_audio_buffer.speech_stopped"
    CONVERSATION_ITEM_CREATED = "conversation.item.created"
    CONVERSATION_ITEM_INPUT_AUDIO_TRANSCRIPTION_COMPLETED = "conversation.item.input_audio_transcription.completed"
    CONVERSATION_ITEM_INPUT_AUDIO_TRANSCRIPTION_FAILED = "conversation.item.input_audio_transcription.failed"
    CONVERSATION_ITEM_TRUNCATED = "conversation.item.truncated"
    CONVERSATION_ITEM_DELETED = "conversation.item.deleted"
    RESPONSE_CREATED = "response.created"
    RESPONSE_DONE = "response.done"
    RESPONSE_OUTPUT_ITEM_ADDED = "response.output_item.added"
    RESPONSE_OUTPUT_ITEM_DONE = "response.output_item.done"
    RESPONSE_CONTENT_PART_ADDED = "response.content_part.added"
    RESPONSE_CONTENT_PART_DONE = "response.content_part.done"
    RESPONSE_TEXT_DELTA = "response.text.delta"
    RESPONSE_TEXT_DONE = "response.text.done"
    RESPONSE_AUDIO_TRANSCRIPT_DELTA = "response.audio_transcript.delta"
    RESPONSE_AUDIO_TRANSCRIPT_DONE = "response.audio_transcript.done"
    RESPONSE_AUDIO_DELTA = "response.audio.delta"
    RESPONSE_AUDIO_DONE = "response.audio.done"
    RESPONSE_FUNCTION_CALL_ARGUMENTS_DELTA = "response.function_call_arguments.delta"
    RESPONSE_FUNCTION_CALL_ARGUMENTS_DONE = "response.function_call_arguments.done"
    RATE_LIMITS_UPDATED = "rate_limits.updated"


class ClientEvent(str, Enum):
    SESSION_UPDATE = "session.update"
    INPUT_AUDIO_BUFFER_APPEND = "input_audio_buffer.append"
    INPUT_AUDIO_BUFFER_COMMIT = "input_audio_buffer.commit"
    INPUT_AUDIO_BUFFER_CLEAR = "input_audio_buffer.clear"
    CONVERSATION_ITEM_CREATE = "conversation.item.create"
    CONVERSATION_ITEM_TRUNCATE = "conversation.item.truncate"
    CONVERSATION_ITEM_DELETE = "conversation.item.delete"
    RESPONSE_CREATE = "response.create"
    RESPONSE_CANCEL = "response.cancel"



import sys

if sys.version_info[:2] < (3, 9):
    raise RuntimeError("This version of Outspeed requires at least Python 3.9")
if sys.version_info[:2] >= (3, 13):
    raise RuntimeError("This version of Outspeed does not support Python 3.13+")


import logging

import coloredlogs


def configure_logging(level=logging.INFO):
    """
    Configure the root logger with a coloredlogs that logs each message to console with a color based on its log level.

    Args:
        level (int): The log level to set for the root logger. Defaults to `logging.INFO`.
    """

    # By default the install() function installs a handler on the root logger,
    # this means that log messages from your code and log messages from the
    # libraries that you use will all show up on the terminal.
    coloredlogs.install(level=level, fmt="%(asctime)s %(levelname)s %(message)s")

    # If you don't want to see log messages from libraries, you can pass a

    # specific logger object to the install() function. In this case only log
    # messages originating from that logger will show up on the terminal.
    # coloredlogs.install(level="DEBUG", logger=logger)


configure_logging()

try:
    from .app import App  # noqa: F401
    from .data import AudioData, ImageData, SessionData, TextData  # noqa: F401
    from .ops.map import map  # noqa: F401
    from .ops.merge import merge  # noqa: F401
    from .plugins.azure_stt import AzureTranscriber  # noqa: F401
    from .plugins.azure_tts import AzureTTS  # noqa: F401
    from .plugins.cartesia_tts import CartesiaTTS  # noqa: F401
    from .plugins.deepgram_stt import DeepgramSTT  # noqa: F401
    from .plugins.eleven_labs_tts import ElevenLabsTTS  # noqa: F401
    from .plugins.fireworks_llm import FireworksLLM  # noqa: F401
    from .plugins.gemini_vision import GeminiVision  # noqa: F401
    from .plugins.groq_llm import GroqLLM  # noqa: F401
    from .plugins.key_frame_detector import KeyFrameDetector  # noqa: F401
    from .plugins.openai_llm import OpenAILLM  # noqa: F401
    from .plugins.openai_realtime.openai_realtime import OpenAIRealtime  # noqa: F401
    from .plugins.openai_vision import OpenAIVision  # noqa: F401
    from .plugins.silero_vad import SileroVAD  # noqa: F401
    from .plugins.token_aggregator import TokenAggregator  # noqa: F401
    from .plugins.whisper_stt import WhisperSTT  # noqa: F401
    from .server import RealtimeServer  # noqa: F401
    from .streaming_endpoint import streaming_endpoint  # noqa: F401
    from .streams import AudioStream, TextStream, VADStream, VideoStream  # noqa: F401
    from .tool import Tool  # noqa: F401
    from .utils.clock import Clock  # noqa: F401
    from .utils.vad import VADState  # noqa: F401
    from .web_endpoint import web_endpoint  # noqa: F401
    from .websocket import websocket  # noqa: F401
except Exception:
    print()
    print("#" * 50)
    print("#" + "Something with the Outspeed installation seems broken.".center(48) + "#")
    print("#" * 50)
    print()
    raise

import av

av.logging.set_level(av.logging.PANIC)


__all__ = [
    "configure_logging",
    "streaming_endpoint",
    "App",
    "web_endpoint",
    "websocket",
    "AudioData",
    "ImageData",
    "TextData",
    "CartesiaTTS",
    "DeepgramSTT",
    "GroqLLM",
    "TokenAggregator",
    "map",
    "merge",
    "AzureTTS",
    "ElevenLabsTTS",
    "FireworksLLM",
    "Clock",
    "SileroVAD",
    "VADStream",
    "SessionData",
    "VADState",
    "WhisperSTT",
    "OpenAILLM",
    "KeyFrameDetector",
    "GeminiVision",
    "OpenAIVision",
    "OpenAIRealtime",
    "Tool",
    "RealtimeServer",
    "AzureTranscriber",
]



from __future__ import annotations

import asyncio
import json
import logging
import os
import websockets
import time
from typing import Any, Dict, Optional, Union
from urllib.parse import urlencode

from outspeed.data import AudioData, SessionData
from outspeed.plugins.base_plugin import Plugin
from outspeed.streams import TextStream
from outspeed.utils import tracing
from pydub import AudioSegment


logger = logging.getLogger(__name__)
SENTENCE_TERMINATORS = [".", "!", "?", "\n", "\r"]
WHISPER_SAMPLE_RATE = 16000


class WhisperSTT(Plugin):
    """
    A plugin for real-time speech-to-text using Deepgram's API.

    This class handles the connection to Deepgram's WebSocket API, sends audio data,
    and processes the returned transcriptions.
    """

    def __init__(
        self,
        *,
        language: str = None,
        detect_language: bool = False,
        interim_results: bool = True,
        punctuate: bool = True,
        smart_format: bool = True,
        model: str = "distil-whisper/distil-small.en",
        api_key: Optional[str] = None,
        sample_rate: int = 16000,
        num_channels: int = 1,
        sample_width: int = 2,
        min_silence_duration: int = 100,
        confidence_threshold: float = 0.8,
        base_url: Optional[str] = None,
        vad_threshold: float = 0.5,
    ) -> None:
        """
        Initialize the DeepgramSTT plugin.

        :param language: The language code for transcription.
        :param detect_language: Whether to enable language detection.
        :param interim_results: Whether to return interim results.
        :param punctuate: Whether to add punctuation to the transcript.
        :param smart_format: Whether to apply smart formatting to the transcript.
        :param model: The Deepgram model to use for transcription.
        :param api_key: The Deepgram API key. If None, it will be read from the DEEPGRAM_API_KEY environment variable.
        :param sample_rate: The sample rate of the audio in Hz.
        :param num_channels: The number of audio channels.
        :param sample_width: The width of each audio sample in bytes.
        :param min_silence_duration: The minimum duration of silence to trigger end of speech, in milliseconds.
        :param confidence_threshold: The minimum confidence score to accept a transcription.
        """
        api_key = api_key or os.getenv("WHISPER_API_KEY")
        if not api_key:
            raise ValueError("Whisper API key is required")

        if not base_url:
            raise ValueError("Whisper base URL is required")

        if sample_rate != 16000 and sample_rate != 8000:
            logging.error(f"Invalid sample rate: {sample_rate}. Whisper sample rate needs to be 16000 or 8000")
            raise ValueError("Whisper sample rate needs to be 16000 or 8000")

        self._api_key: str = api_key
        self.base_url: str = base_url
        self.language: str = language
        self.detect_language: bool = detect_language
        self.interim_results: bool = interim_results
        self.punctuate: bool = punctuate
        self.vad_threshold: float = vad_threshold
        self.smart_format: bool = smart_format
        self.model: str = model
        self.min_silence_duration: int = min_silence_duration
        self.endpointing: int = min_silence_duration

        self._sample_rate: int = sample_rate
        self._num_channels: int = num_channels
        self._sample_width: int = sample_width
        self._speaking: bool = False
        self.confidence_threshold: float = confidence_threshold

        self._closed: bool = False
        self.output_queue: TextStream = TextStream()
        self._audio_duration_received: float = 0.0
        self.input_queue: Optional[asyncio.Queue] = None
        self._task: Optional[asyncio.Task] = None
        self._ws: Optional[websockets.WebSocketClientProtocol] = None
        self._whisper_chunk_size = 512 if self._sample_rate == 16000 else 256
        self._whisper_chunk_bytes = self._whisper_chunk_size * self._num_channels * 2

    def resample_audio(self, audio_bytes: bytes) -> bytes:
        if self._sample_rate == 16000:
            return audio_bytes

        audio_segment = AudioSegment(
            data=bytes(audio_bytes),
            sample_width=2,  # Assuming 16-bit audio
            frame_rate=self._sample_rate,  # Whisper sample rate
            channels=1,
        )

        # Resample the audio to 16000 Hz
        resampled_audio = audio_segment.set_frame_rate(WHISPER_SAMPLE_RATE)

        # Convert the resampled audio back to bytes
        data = resampled_audio.raw_data
        if len(data) < 1024:
            data += b"\0" * (1024 - len(data))

        return data

    def run(self, input_queue: asyncio.Queue) -> TextStream:
        """
        Start the Deepgram STT process.

        :param input_queue: The queue to receive audio data from.
        :return: The output queue for transcribed text.
        """
        self.input_queue = input_queue
        self._task = asyncio.create_task(self._run_ws())
        return self.output_queue

    async def _connect_ws(self) -> None:
        """Connect to the Deepgram WebSocket API."""
        live_config: Dict[str, Any] = {
            "model_name": self.model,
            "sample_rate": WHISPER_SAMPLE_RATE,  # Whisper sample rate
            "channels": self._num_channels,
            "language": self.language,
            "min_silence_ms": self.min_silence_duration,
            "thresh": self.vad_threshold,
        }

        # headers = {"Authorization": f"Token {self._api_key}"}

        # url = f"{self.base_url}?{urlencode(live_config).lower()}"
        url = self.base_url
        try:
            socket = await websockets.connect(url)
            await socket.send(json.dumps(live_config))
            self._ws = socket
        except Exception:
            logger.error("Whisper connection failed", exc_info=True)
            raise asyncio.CancelledError()

    async def _run_ws(self) -> None:
        """Run the main WebSocket communication loop with Deepgram."""
        try:
            await asyncio.gather(self._send_task(), self._recv_task())
        except Exception:
            logger.error("Whisper task failed", exc_info=True)

    async def _send_task(self) -> None:
        """
        Send audio data to Deepgram through the WebSocket.

        :param ws: The WebSocket connection to Deepgram.
        """
        try:
            buffer = b""
            while True:
                data: Union[AudioData, SessionData] = await self.input_queue.get()
                if not self._ws:
                    await self._connect_ws()

                if isinstance(data, SessionData):
                    await self.output_queue.put(data)
                    continue

                bytes_data = data.get_bytes()
                buffer += bytes_data
                self._audio_duration_received += len(bytes_data) / (
                    self._sample_rate * self._num_channels * self._sample_width
                )
                while len(buffer) >= self._whisper_chunk_bytes:
                    resampled_buffer = self.resample_audio(buffer[: self._whisper_chunk_bytes])
                    await self._ws.send(resampled_buffer)
                    buffer = buffer[self._whisper_chunk_bytes :]
        except Exception:
            logger.error("Whisper send task failed", exc_info=True)
            raise asyncio.CancelledError()

    async def _recv_task(self) -> None:
        """
        Receive and process transcription results from Deepgram.

        :param ws: The WebSocket connection to Deepgram.
        """
        try:
            buffer = ""
            while True:
                if not self._ws:
                    await asyncio.sleep(0.2)
                    continue

                try:
                    msg = await asyncio.wait_for(self._ws.recv(), timeout=1)
                except asyncio.TimeoutError:
                    if buffer:
                        await self.output_queue.put(buffer)
                        buffer = ""
                    continue
                logger.info(f"Whisper msg: {msg}")

                if not msg:
                    continue

                if msg[-1] in SENTENCE_TERMINATORS:
                    await self.output_queue.put(buffer + msg)
                    buffer = ""
                else:
                    buffer += msg
        except Exception:
            logger.error("Whisper receive task failed", exc_info=True)
            raise asyncio.CancelledError()



import asyncio
import logging
import os
from typing import Any, Callable, Optional, Type

from dotenv import load_dotenv

from outspeed._realtime_function import RealtimeFunction
from outspeed.server import RealtimeServer


def App(dotenv_path: Optional[str] = ".env") -> Callable[[Type], Callable[..., "RealtimeApp"]]:
    """
    Decorator factory for creating a RealtimeApp.

    Args:
        dotenv_path (str | None): The path to the .env file to load variables from. Defaults to ".env".

    Returns:
        A decorator function that wraps a user-defined class.
    """

    if dotenv_path and os.path.exists(dotenv_path):
        logging.info(f"Loading variables from {dotenv_path} file...")
        load_dotenv(dotenv_path=dotenv_path, override=True)

    def wrapper(user_cls: Type) -> Callable[..., "RealtimeApp"]:
        def construct(*args: Any, **kwargs: Any) -> "RealtimeApp":
            return RealtimeApp(user_cls=user_cls, *args, **kwargs)

        return construct

    return wrapper


class RealtimeApp:
    """
    Main application class for the Realtime framework.
    """

    functions: list = []  # List to store realtime functions (currently unused)

    def __init__(self, user_cls: Type, *args: Any, **kwargs: Any):
        """
        Initialize the RealtimeApp.

        Args:
            user_cls: The user-defined class to be wrapped.
            *args: Variable length argument list.
            **kwargs: Arbitrary keyword arguments.
        """
        self._user_cls: Type = user_cls
        self._user_cls_instance: Any = self._user_cls(*args, **kwargs)

    def __getattr__(self, k: str) -> Any:
        """
        Attribute lookup method.

        Args:
            k: The name of the attribute to look up.

        Returns:
            The value of the attribute.

        Raises:
            AttributeError: If the attribute is not found.
        """
        if hasattr(self, k):
            return getattr(self, k)
        elif hasattr(self._user_cls_instance, k):
            return getattr(self._user_cls_instance, k)
        else:
            raise AttributeError(k)

    def start(self) -> None:
        """
        Start the Realtime application.

        This method sets up the event loop, runs the setup, main loop, and teardown methods
        of the user-defined class, and handles the RealtimeServer.
        """
        try:
            loop = asyncio.get_event_loop()
        except RuntimeError:
            loop = asyncio.new_event_loop()

        rt_functions = RealtimeFunction.get_realtime_functions_from_class(self._user_cls_instance)
        if len(rt_functions) > 1:
            raise RuntimeError("More than one realtime function found in the user class.")
        try:
            # Run setup
            loop.run_until_complete(self._user_cls_instance.setup())

            # Run main loop and RealtimeServer concurrently
            if len(rt_functions) == 1:
                rt_func = list(rt_functions.values())[0]
                loop.run_until_complete(asyncio.gather(RealtimeServer().start(), rt_func(self._user_cls_instance)))
            else:
                loop.run_until_complete(asyncio.gather(RealtimeServer().start()))
        except asyncio.CancelledError:
            logging.error("RealtimeServer was cancelled")
        except Exception as e:
            logging.error(f"Error in RealtimeApp: {e}")

        finally:
            # Run teardown
            try:
                loop.run_until_complete(self._user_cls_instance.teardown())
            except Exception as e:
                pass



import asyncio
import json
import logging
import os
from typing import Tuple

from openai import AsyncOpenAI

from outspeed.plugins.base_plugin import Plugin
from outspeed.streams import TextStream, VADStream
from outspeed.utils import tracing

from outspeed.utils.vad import VADState
from outspeed.data import SessionData


class FireworksLLM(Plugin):
    def __init__(
        self,
        model: str = "accounts/fireworks/models/llama-v3p1-8b-instruct",
        api_key=None,
        base_url=None,
        system_prompt=None,
        stream: bool = True,
        temperature: float = 1.0,
        response_format: dict = None,
    ):
        super().__init__()
        self._model: str = model
        api_key = api_key or os.getenv("FIREWORKS_API_KEY")
        if not api_key:
            raise ValueError("Fireworks API key is required")
        self._api_key = api_key
        self._client = AsyncOpenAI(api_key=api_key, base_url="https://api.fireworks.ai/inference/v1")
        self._history = []
        self.output_queue = TextStream()
        self.chat_history_queue = TextStream()
        self._generating = False
        self._stream = stream
        self._response_format = response_format
        self._system_prompt = system_prompt
        if self._system_prompt is not None:
            self._history.append({"role": "system", "content": self._system_prompt})
        self._temperature = temperature

    async def _stream_chat_completions(self):
        try:
            while True:
                text_chunk = await self.input_queue.get()
                if text_chunk is None:
                    continue

                if isinstance(text_chunk, SessionData):
                    await self.output_queue.put(text_chunk)
                    continue
                self._generating = True
                self._history.append({"role": "user", "content": text_chunk})
                self.chat_history_queue.put_nowait(json.dumps(self._history[-1]))
                tracing.register_event(tracing.Event.LLM_START)
                if self._response_format:
                    chunk_stream = await self._client.chat.completions.create(
                        model=self._model,
                        stream=self._stream,
                        messages=self._history,
                        response_format=self._response_format,
                        temperature=self._temperature,
                    )
                else:
                    chunk_stream = await self._client.chat.completions.create(
                        model=self._model,
                        stream=self._stream,
                        messages=self._history,
                        temperature=self._temperature,
                    )
                self._history.append({"role": "assistant", "content": ""})
                tracing.register_event(tracing.Event.LLM_TTFB)
                if self._stream:
                    async for chunk in chunk_stream:
                        if len(chunk.choices) == 0:
                            continue

                        elif chunk.choices[0].delta.content:
                            self._history[-1]["content"] += chunk.choices[0].delta.content
                            await self.output_queue.put(chunk.choices[0].delta.content)
                else:
                    self._history[-1]["content"] = chunk_stream.choices[0].message.content
                    await self.output_queue.put(chunk_stream.choices[0].message.content)
                tracing.register_event(tracing.Event.LLM_END)
                tracing.register_metric(tracing.Metric.LLM_TOTAL_BYTES, len(self._history[-1]["content"]))
                logging.info("llm: %s", self._history[-1]["content"])
                self.chat_history_queue.put_nowait(json.dumps(self._history[-1]))
                self._generating = False
                await self.output_queue.put(None)
        except Exception as e:
            logging.error("Error streaming chat completions", e)
            self._generating = False
            raise asyncio.CancelledError()

    def run(self, input_queue: TextStream) -> Tuple[TextStream, TextStream]:
        self.input_queue = input_queue
        self._task = asyncio.create_task(self._stream_chat_completions())
        return self.output_queue, self.chat_history_queue

    async def close(self):
        self._task.cancel()

    async def _interrupt(self):
        while True:
            vad_state: VADState = await self.interrupt_queue.get()
            if vad_state == VADState.SPEAKING and (not self.input_queue.empty() or not self.output_queue.empty()):
                self._task.cancel()
                try:
                    await self._task
                except asyncio.CancelledError:
                    pass
                while not self.output_queue.empty():
                    self.output_queue.get_nowait()
                while not self.input_queue.empty():
                    self.input_queue.get_nowait()
                logging.info("Done cancelling LLM")
                self._generating = False
                self._task = asyncio.create_task(self._stream_chat_completions())

    def set_interrupt_stream(self, interrupt_stream: VADStream):
        if isinstance(interrupt_stream, VADStream):
            self.interrupt_queue = interrupt_stream
        else:
            raise ValueError("Interrupt stream must be a VADStream")
        self._interrupt_task = asyncio.create_task(self._interrupt())



import asyncio
import functools
import inspect
import logging
import traceback
from typing import Callable, List, Optional, Tuple, Union

from outspeed._realtime_function import RealtimeFunction
from outspeed.server import RealtimeServer
from outspeed.streaming_endpoint.AudioRTCDriver import AudioRTCDriver
from outspeed.streaming_endpoint.server import create_and_run_server
from outspeed.streaming_endpoint.TextRTCDriver import TextRTCDriver
from outspeed.streaming_endpoint.VideoRTCDriver import VideoRTCDriver
from outspeed.streams import AudioStream, TextStream, VideoStream
from outspeed.utils import tracing
from outspeed.utils.audio import AudioCodec
from outspeed.utils.images import VideoCodec

logger = logging.getLogger(__name__)


def streaming_endpoint(audio_codec: str = AudioCodec.OPUS, video_codec: str = VideoCodec.H264) -> Callable:
    """
    Decorator for creating a streaming endpoint.

    This decorator wraps a function to set up and manage audio, video, and text streams
    for real-time communication.

    Returns:
        Callable: A decorator function.
    """

    def decorator(func: Callable) -> Callable:
        @functools.wraps(func)
        async def wrapper(*args, **kwargs) -> None:
            try:
                # Initialize input queues
                audio_input_q: Optional[AudioStream] = None
                video_input_q: Optional[VideoStream] = None
                text_input_q: Optional[TextStream] = None

                # Inspect the function signature and set up input streams
                signature = inspect.signature(func)
                parameters = signature.parameters
                for name, param in parameters.items():
                    if param.annotation == AudioStream:
                        audio_input_q = AudioStream()
                        kwargs[name] = audio_input_q
                    elif param.annotation == VideoStream:
                        video_input_q = VideoStream()
                        kwargs[name] = video_input_q
                    elif param.annotation == TextStream:
                        text_input_q = TextStream()
                        kwargs[name] = text_input_q

                tracing.start()
                # Call the wrapped function and get output streams
                output_streams = await func(*args, **kwargs)
                # Ensure output_streams is iterable
                if not isinstance(output_streams, (list, tuple)):
                    output_streams = (output_streams,)

                # Initialize output queues
                aq, vq, tq = None, None, None
                for s in output_streams:
                    if isinstance(s, AudioStream):
                        aq = s
                    elif isinstance(s, VideoStream):
                        vq = s
                    elif isinstance(s, TextStream):
                        tq = s

                # Set up RTC drivers for each stream type
                video_output_frame_processor = VideoRTCDriver(video_input_q, vq)
                # TODO: get audio_output_layout, audio_output_format, audio_output_sample_rate from SDP
                audio_output_frame_processor = AudioRTCDriver(audio_input_q, aq)
                text_output_processor = TextRTCDriver(text_input_q, tq)

                # Create and run the server
                create_and_run_server(
                    audio_output_frame_processor,
                    video_output_frame_processor,
                    text_output_processor,
                    audio_codec,
                    video_codec,
                )

                # Create and run tasks for each processor
                tasks = [
                    asyncio.create_task(video_output_frame_processor.run_input()),
                    asyncio.create_task(audio_output_frame_processor.run()),
                    asyncio.create_task(text_output_processor.run_input()),
                ]
                await asyncio.gather(*tasks)

            except asyncio.CancelledError:
                logging.error("streaming_endpoint: CancelledError")
            except Exception as e:
                logging.error(f"Error in streaming_endpoint: {e}")
                logging.error(f"Traceback:\n{traceback.format_exc()}")
            finally:
                RealtimeServer().remove_connection()
                tracing.end()
                logging.info("Received exit, stopping bot")
                # Clean up tasks
                loop = asyncio.get_event_loop()
                tasks = asyncio.all_tasks(loop)
                for task in tasks:
                    try:
                        task.cancel()
                        await task
                    except asyncio.CancelledError:
                        pass

        rt_func = RealtimeFunction(wrapper)
        return rt_func

    return decorator



import os

import click
import dill
import requests


@click.group()
@click.version_option(message="%(prog)s version %(version)s", help="Show the version of the SDK.")
def cli():
    """A CLI tool for deploying serialized Python files."""
    pass


@cli.command()  # This adds the deploy function as a sub-command of outspeed
@click.argument(
    "file_path",
    type=click.Path(exists=True),
    required=True,
    metavar="FILE_PATH",
)
@click.option("--api-key", required=False, help="API key of the sending user")
@click.option("--base-url", required=False, help="Base URL of Outspeed's endpoint")
def deploy(file_path, api_key, base_url):
    """Serializes a .py file and sends it to the specified backend server."""
    BASE_URL = base_url or "https://infra.outspeed.com"
    endpoint = f"{BASE_URL}/deploy"
    api_key = api_key or os.getenv("OUTSPEED_API_KEY")
    if not api_key:
        click.echo(
            "No API key provided. Please set the OUTSPEED_API_KEY environment variable or use the --api-key flag."
        )
        return
    try:
        with open(file_path, "rb") as file:
            fmodule_content = file.read()

            payload = {
                "module": fmodule_content,
                "metadata": {
                    # send the file name as the function name
                    "name": os.path.basename(file_path),
                },
            }
            serialized_payload = dill.dumps(payload, recurse=True)
            headers = {"Content-Type": "application/octet-stream", "X-API-KEY": api_key}
            response = None
            try:
                response = requests.post(endpoint, data=serialized_payload, headers=headers)
            except Exception as e:
                click.echo(f"An error occurred while connecting to the server: {str(e)}")
                click.echo(f"Base URL: {BASE_URL}, Headers: {headers}")
                if response:
                    click.echo(f"Response: {response.text}")
                return
            try:
                body = response.json()
            except Exception as e:
                click.echo(f"An error occurred while parsing the response: {str(e)}")
                click.echo(f"Response: {response.text}")
                return
            if response.status_code == 200:
                click.echo("Function successfully deployed!")
                click.echo(f"Use function URL: {BASE_URL}/run/{body['functionId']} to run the function")
            else:
                click.echo(f"Failed to deploy file. Status code: {response.status_code}")
                click.echo(f"Response: {response.text}")
    except Exception as e:
        click.echo(f"An error occurred: {str(e)}")


if __name__ == "__main__":
    cli()



import asyncio
import json
import logging
import os
import time
from typing import Optional

import google.generativeai as genai
import PIL.PngImagePlugin  # Not used but needed to make Gemini API work with PIL  # noqa: F401

from outspeed.data import SessionData
from outspeed.plugins.base_plugin import Plugin
from outspeed.streams import TextStream, VideoStream, VADStream
from outspeed.data import ImageData
from outspeed.utils.vad import VADState

logger = logging.getLogger(__name__)


class GeminiVision(Plugin):
    def __init__(
        self,
        model: str = "gemini-1.5-flash-latest",
        api_key: Optional[str] = None,
        base_url: Optional[str] = None,
        system_prompt: Optional[str] = None,
        temperature: float = 1.0,
        store_image_history: bool = True,
        stream: bool = True,
        max_output_tokens: int = 75,
    ):
        super().__init__()
        self._model: str = model

        api_key = api_key or os.getenv("GOOGLE_API_KEY")
        if not api_key:
            raise ValueError("GOOGLE API key is required")

        genai.configure(api_key=api_key)
        self._client = genai.GenerativeModel(model)
        self._history = []
        self._system_prompt = system_prompt
        self._temperature = temperature
        self.chat_history_queue = TextStream()
        if self._system_prompt is not None:
            self._history.append({"role": "user", "parts": [self._system_prompt]})
        self.output_queue = TextStream()
        self._store_image_history = store_image_history
        self._stream = stream
        self._max_output_tokens = max_output_tokens
        self._safety_settings = [
            {
                "category": "HARM_CATEGORY_HARASSMENT",
                "threshold": "BLOCK_NONE",
            },
            {
                "category": "HARM_CATEGORY_HATE_SPEECH",
                "threshold": "BLOCK_NONE",
            },
            {
                "category": "HARM_CATEGORY_SEXUALLY_EXPLICIT",
                "threshold": "BLOCK_NONE",
            },
            {
                "category": "HARM_CATEGORY_DANGEROUS_CONTENT",
                "threshold": "BLOCK_NONE",
            },
        ]

    async def _stream_chat_completions(self):
        try:
            while True:
                prompt = await self.input_queue.get()
                if prompt is None:
                    continue

                if isinstance(prompt, SessionData):
                    await self.output_queue.put(prompt)
                    continue

                self._generating = True
                start_time = time.time()
                self._history.append(
                    {
                        "role": "user",
                        "parts": [],
                    }
                )
                if isinstance(prompt, ImageData):
                    logger.info("GeminiVision prompt: %s", "image")
                    content = prompt.get_pil_image()
                    self._history[-1]["parts"].append(content)
                    self.chat_history_queue.put_nowait(
                        json.dumps(
                            {
                                "role": "user",
                                "content": "Image",
                            }
                        )
                    )
                else:
                    logger.info("GeminiVision prompt: %s", prompt)
                    content = prompt
                    self._history[-1]["parts"].append(content)
                    self.chat_history_queue.put_nowait(
                        json.dumps(
                            {
                                "role": "user",
                                "content": content,
                            }
                        )
                    )

                try:
                    response = await self._client.generate_content_async(
                        self._history,
                        stream=self._stream,
                        generation_config=genai.types.GenerationConfig(
                            max_output_tokens=self._max_output_tokens, temperature=self._temperature
                        ),
                        safety_settings=self._safety_settings,
                    )
                except Exception as e:
                    logger.error("Google AI vision timeout %s", e)
                    self._history.pop()
                    continue

                # if prompt != "":
                #     self._history[-1]["parts"] = self._history[-1]["parts"][:1]
                # else:
                if not self._store_image_history:
                    self._history.pop()
                self._history.append({"role": "model", "parts": [""]})
                logger.info("Google AI LLM TTFB: %s", time.time() - start_time)
                if self._stream:
                    async for chunk in response:
                        if chunk:
                            try:
                                text = chunk.text
                            except:
                                continue
                            self._history[-1]["parts"][0] += text
                            await self.output_queue.put(text)
                else:
                    text = response.text
                    self._history[-1]["parts"][0] += text
                    await self.output_queue.put(text)
                logger.info("llm %s", self._history[-1]["parts"][0])
                self.chat_history_queue.put_nowait(
                    json.dumps(
                        {
                            "role": "assistant",
                            "content": self._history[-1]["parts"][0],
                        }
                    )
                )
                if not self._store_image_history:
                    self._history.pop()
                self._generating = False
                await self.output_queue.put(None)
        except Exception as e:
            logger.error("GeminiVision error %s", e)
            raise asyncio.CancelledError()

    def run(self, input_queue: VideoStream) -> TextStream:
        self.input_queue = input_queue
        self._task = asyncio.create_task(self._stream_chat_completions())
        return self.output_queue, self.chat_history_queue

    async def close(self):
        self._task.cancel()

    async def _interrupt(self):
        while True:
            vad_state: VADState = await self.interrupt_queue.get()
            if vad_state == VADState.SPEAKING and (
                not self.input_queue.empty() or not self.output_queue.empty() or self._generating
            ):
                self._task.cancel()
                try:
                    await self._task
                except asyncio.CancelledError:
                    pass
                while not self.output_queue.empty():
                    self.output_queue.get_nowait()
                while not self.input_queue.empty():
                    self.input_queue.get_nowait()
                logging.info("Done cancelling LLM")
                self._generating = False
                self._task = asyncio.create_task(self._stream_chat_completions())

    def set_interrupt_stream(self, interrupt_stream: VADStream):
        if isinstance(interrupt_stream, VADStream):
            self.interrupt_queue = interrupt_stream
        else:
            raise ValueError("Interrupt stream must be a VADStream")
        self._interrupt_task = asyncio.create_task(self._interrupt())



import logging
import os
import time

import numpy as np
import torch

# Brute force torchaudio to use ffmpeg 7, otherwise it will use ffmpeg 6 which causes issues on mac
os.environ["TORIO_USE_FFMPEG_VERSION"] = "7"
# Although torchaudio is not utilized directly in this context, it is necessary to attempt its import as it is a dependency for Silero.
import torchaudio  # noqa: F401
from silero_vad import load_silero_vad


class SileroVADModel:
    """
    A Voice Activity Detection (VAD) model using Silero VAD.

    This class implements a VAD model that uses the Silero VAD model to detect voice activity
    in audio samples.

    Attributes:
        _sample_rate (int): The sample rate of the audio input.
        _num_channels (int): The number of audio channels.
        _model: The loaded Silero VAD model.
        _last_reset_time (float): The timestamp of the last model state reset.
        _model_reset_states_time (float): The time interval for resetting model states.
    """

    def __init__(
        self,
        *,
        sample_rate: int,
        num_channels: int,
        model_reset_states_time: float = 5.0,
    ):
        """
        Initialize the SileroVADModel.

        Args:
            sample_rate (int): The sample rate of the audio input (must be 16000 or 8000).
            num_channels (int): The number of audio channels.
            model_reset_states_time (float): The time interval for resetting model states.

        Raises:
            ValueError: If the sample rate is not 16000 or 8000.
        """
        self._sample_rate = sample_rate

        if sample_rate != 16000 and sample_rate != 8000:
            logging.error(f"Invalid sample rate: {sample_rate}. Silero VAD sample rate needs to be 16000 or 8000")
            raise ValueError("Silero VAD sample rate needs to be 16000 or 8000")

        self._num_channels = num_channels

        logging.debug(f"Initializing SileroVADModel with sample rate: {sample_rate}, channels: {num_channels}")
        # self._model, _ = torch.hub.load(repo_or_dir="snakers4/silero-vad", model="silero_vad", force_reload=False)
        self._model = load_silero_vad()
        logging.debug("Silero VAD model loaded successfully")

        self._last_reset_time = 0
        self._model_reset_states_time = model_reset_states_time

    def voice_confidence(self, buffer: bytes) -> float:
        """
        Calculate the voice confidence for the given audio buffer.

        Args:
            buffer (bytes): The audio buffer to analyze.

        Returns:
            float: The voice confidence score (0.0 to 1.0).
        """
        try:
            audio_int16 = np.frombuffer(buffer, np.int16)
            # Divide by 32768 because we have signed 16-bit data.
            audio_float32 = np.frombuffer(audio_int16, dtype=np.int16).astype(np.float32) / 32768.0
            new_confidence = self._model(torch.from_numpy(audio_float32), self._sample_rate).item()

            # We need to reset the model from time to time because it doesn't
            # really need all the data and memory will keep growing otherwise.
            curr_time = time.time()
            diff_time = curr_time - self._last_reset_time
            if diff_time >= self._model_reset_states_time:
                logging.debug(f"Resetting Silero VAD model states after {diff_time:.2f} seconds")
                self._model.reset_states()
                self._last_reset_time = curr_time

            logging.debug(f"Voice confidence: {new_confidence:.4f}")
            return new_confidence
        except Exception as e:
            # This comes from an empty audio array
            logging.error(f"Error analyzing audio with Silero VAD: {e}")
            return 0



import time


class Clock:
    start_time = None

    @classmethod
    def start_playback(cls):
        if cls.start_time is None:
            cls.start_time = time.time()

    @classmethod
    def get_playback_time(cls) -> float:
        cls.start_playback()
        return time.time() - cls.start_time



import asyncio
from typing import Any, Callable, Dict, Type


class RealtimeFunction:
    def __init__(
        self,
        raw_f: Callable[..., Any],
    ):
        self.raw_f = raw_f
        self.is_async = asyncio.iscoroutinefunction(raw_f)

    def __call__(self, *args: Any, **kwargs: Any) -> Any:
        return self.raw_f(*args, **kwargs)

    @classmethod
    def get_realtime_functions_from_class(cls, user_cls: Type):
        realtime_functions: Dict[str, RealtimeFunction] = {}
        for name in dir(user_cls):
            if name.startswith("__"):  # Skip magic methods
                continue
            attr = getattr(user_cls, name)
            if isinstance(attr, RealtimeFunction):
                realtime_functions[name] = attr
        return realtime_functions



import asyncio

from outspeed.streams import AudioStream, ByteStream, Stream, TextStream, VideoStream


def unzip_array(input_queue: Stream):
    output_queue = None
    if isinstance(input_queue, AudioStream):
        output_queue = AudioStream()
    elif isinstance(input_queue, VideoStream):
        output_queue = VideoStream()
    elif isinstance(input_queue, TextStream):
        output_queue = TextStream()
    elif isinstance(input_queue, ByteStream):
        output_queue = ByteStream()
    else:
        raise ValueError(f"Invalid input queue type: {type(input_queue)}")

    async def run():
        while True:
            item = await input_queue.get()
            try:
                if isinstance(item, list):
                    for i in item:
                        await output_queue.put(i)
            except Exception as e:
                print(f"Error in unzip array function: {e}")
                continue

    asyncio.create_task(run())
    return output_queue



import asyncio
import base64
import logging
import time

import numpy as np
import scipy.signal as signal
from fastapi import WebSocket

from outspeed.data import AudioData
from outspeed.streams import AudioStream, ByteStream, TextStream, VideoStream


def resample_wav_bytes(audio_data: AudioData, target_sample_rate: int) -> bytes:
    """
    Resample WAV bytes to a target sample rate.

    Args:
        wav_bytes (bytes): The input WAV file as bytes.
        target_sample_rate (int): The desired sample rate in Hz.

    Returns:
        bytes: The resampled WAV file as bytes.
    """
    wav_bytes = audio_data.get_bytes()
    if audio_data.sample_rate == target_sample_rate:
        return wav_bytes
    # Load WAV bytes into AudioSegment
    audio_array = np.frombuffer(wav_bytes, dtype=np.int16)

    # Calculate the resampling ratio
    ratio = target_sample_rate / audio_data.sample_rate

    # Resample the audio using scipy.signal.resample
    resampled_audio = signal.resample(audio_array, int(len(audio_array) * ratio))

    resampled_audio = resampled_audio.astype(np.int16).tobytes()

    return resampled_audio


class WebsocketInputProcessor:
    """
    Handles incoming WebSocket messages and streams audio and text data.

    Attributes:
        sample_rate (int): The sample rate for audio processing.
        ws (WebSocket): The WebSocket connection.
        audio_output_stream (AudioStream): The stream for input audio data.
        message_stream (TextStream): The stream for input text messages.
        video_stream (VideoStream): The stream for input video data.
    """

    @property
    def sample_rate(self) -> int:
        return self._sample_rate

    @sample_rate.setter
    def sample_rate(self, value: int):
        if not isinstance(value, int) or value <= 0:
            raise ValueError("Sample rate must be a positive integer")
        self._sample_rate = value

    def __init__(
        self, audio_stream: AudioStream, message_stream: TextStream, video_stream: VideoStream, sample_rate: int = 48000
    ):
        self.audio_output_stream = audio_stream
        self.message_stream = message_stream
        self.video_stream = video_stream
        self.sample_rate = sample_rate
        self._inputTrack = None

    def setInputTrack(self, track: TextStream):
        self._inputTrack = track

    async def run(self):
        """
        Starts the task to process incoming WebSocket messages.

        Returns:
            Tuple[AudioStream, TextStream]: A tuple containing the audio and message streams.
        """
        # TODO: Implement video stream processing

        while not self._inputTrack:
            await asyncio.sleep(0.2)
        audio_data = b""
        while True:
            try:
                data = await self._inputTrack.get()
                if data.get("type") == "message":
                    await self.message_stream.put(data.get("data"))
                elif data.get("type") == "audio":
                    audio_bytes = base64.b64decode(data.get("data"))
                    audio_data = AudioData(audio_bytes, sample_rate=self.sample_rate)
                    await self.audio_output_stream.put(audio_data)
            except Exception as e:
                logging.error("websocket: Exception", e)
                raise asyncio.CancelledError()


class WebsocketOutputProcessor:
    """
    Handles outgoing WebSocket messages by streaming audio and text data.

    Attributes:
        sample_rate (int): The sample rate for audio processing.
        audio_stream (AudioStream): The audio stream to send.
        message_stream (TextStream): The text stream to send.
        video_stream (VideoStream): The video stream to send.
        byte_stream (ByteStream): The byte stream to send.
    """

    @property
    def sample_rate(self) -> int:
        return self._sample_rate

    @sample_rate.setter
    def sample_rate(self, value: int):
        if not isinstance(value, int) or value <= 0:
            raise ValueError("Sample rate must be a positive integer")
        self._sample_rate = value

    def __init__(
        self, audio_stream: AudioStream, message_stream: TextStream, video_stream: VideoStream, byte_stream: ByteStream
    ):
        self.audio_stream = audio_stream
        self.message_stream = message_stream
        self.video_stream = video_stream
        self.byte_stream = byte_stream
        self._outputTrack = None

    def setOutputTrack(self, track: TextStream):
        self._outputTrack = track

    async def run(self):
        """
        Starts tasks to process and send byte and text streams.
        """
        # TODO: Implement video stream and audio stream processing
        await asyncio.gather(self.task(self.byte_stream), self.task(self.message_stream), self.task(self.audio_stream))

    async def task(self, input_stream):
        """
        Sends data from the input stream over the WebSocket.

        Args:
            input_stream (Stream): The stream from which to send data.
        """
        while not self._outputTrack:
            await asyncio.sleep(0.2)
        while True:
            if not input_stream:
                break
            audio_data = await input_stream.get()
            if audio_data is None:
                print("Sending audio end")
                json_data = {"type": "audio_end", "timestamp": time.time()}
                await self._outputTrack.put(json_data)
            elif isinstance(audio_data, AudioData):
                data = resample_wav_bytes(audio_data, self.sample_rate)
                json_data = {
                    "type": "audio",
                    "data": base64.b64encode(data).decode(),
                    "timestamp": time.time(),
                    "sample_rate": audio_data.sample_rate,
                }
                await self._outputTrack.put(json_data)
            elif isinstance(audio_data, str):
                json_data = {"type": "message", "data": audio_data, "timestamp": time.time()}
                await self._outputTrack.put(json_data)
            else:
                raise ValueError(f"Unsupported data type: {type(audio_data)}")



import asyncio
import fractions
import logging
import time

from aiortc import MediaStreamTrack
from av import AudioResampler

from outspeed.data import AudioData, SessionData


class AudioRTCDriver(MediaStreamTrack):
    kind = "audio"

    def __init__(
        self,
        audio_input_q,
        audio_output_q,
        output_audio_sample_rate=48000,
        output_audio_layout="stereo",
        output_audio_format="s16",
    ):
        super().__init__()
        self.audio_input_q = audio_input_q
        self.audio_output_q = audio_output_q
        self.audio_data_q = asyncio.Queue()
        self.audio_samples = 0
        self._start = None
        self._track = None
        self.output_audio_sample_rate = output_audio_sample_rate
        self.output_audio_layout = output_audio_layout
        self.output_audio_format = output_audio_format
        self.output_audio_time_base = fractions.Fraction(1, self.output_audio_sample_rate)
        self.output_audio_chunk_size_seconds = 0.020
        self.output_audio_resampler = AudioResampler(
            format=self.output_audio_format,
            layout=self.output_audio_layout,
            rate=self.output_audio_sample_rate,
            frame_size=int(self.output_audio_sample_rate * self.output_audio_chunk_size_seconds),
        )

    async def recv(self):
        frame = await self.audio_data_q.get()
        data_time = frame.samples / frame.sample_rate
        if self._start is None:
            self._start = time.time() + data_time
        else:
            wait = self._start - time.time() - data_time
            if wait > 0:
                await asyncio.sleep(wait)
            self._start = max(self._start, time.time()) + data_time
        return frame

    async def run_input(self):
        try:
            if not self.audio_input_q:
                return
            while not self._track:
                await asyncio.sleep(0.2)
            await self.audio_input_q.put(SessionData())
            while True:
                frame = await self._track.recv()
                await self.audio_input_q.put(AudioData(frame))
        except Exception as e:
            logging.error(f"Error in audio_frame_callback: {str(e)}")
            raise asyncio.CancelledError

    async def run_output(self):
        try:
            while True:
                while self.audio_data_q.qsize() > 2:
                    await asyncio.sleep(0.01)
                audio_data: AudioData = await self.audio_output_q.get()
                if audio_data is None or not isinstance(audio_data, AudioData):
                    continue
                self.audio_samples = max(self.audio_samples, audio_data.get_pts())
                for nframe in self.output_audio_resampler.resample(audio_data.get_frame()):
                    # fix timestamps
                    nframe.pts = self.audio_samples
                    nframe.time_base = self.output_audio_time_base
                    self.audio_samples += nframe.samples
                    self.audio_data_q.put_nowait(nframe)
        except Exception as e:
            logging.error(f"Error in audio_frame_callback: {str(e)}")
            raise asyncio.CancelledError

    async def run(self):
        await asyncio.gather(self.run_input(), self.run_output())

    def add_track(self, track):
        self._track = track



from enum import Enum

import numpy as np
import pyloudnorm as pyln


class AudioCodec(str, Enum):
    OPUS = "audio/opus"
    PCMU = "audio/PCMU"
    PCMA = "audio/PCMA"


def normalize_value(value, min_value, max_value):
    normalized = (value - min_value) / (max_value - min_value)
    normalized_clamped = max(0, min(1, normalized))
    return normalized_clamped


def calculate_audio_volume(audio: bytes, sample_rate: int) -> float:
    audio_np = np.frombuffer(audio, dtype=np.int16)
    audio_float = audio_np.astype(np.float64)

    block_size = audio_np.size / sample_rate
    meter = pyln.Meter(sample_rate, block_size=block_size)
    loudness = meter.integrated_loudness(audio_float)

    # Loudness goes from -20 to 80 (more or less), where -20 is quiet and 80 is
    # loud.
    loudness = normalize_value(loudness, -20, 80)

    return loudness


def exp_smoothing(value: float, prev_value: float, factor: float) -> float:
    return prev_value + factor * (value - prev_value)



import asyncio
import json
import logging
import os
from typing import Any, Dict, List, Optional, Tuple

from openai import AsyncOpenAI

from outspeed.plugins.base_plugin import Plugin
from outspeed.streams import TextStream
from outspeed.utils import tracing
from outspeed.data import SessionData
from outspeed.streams import VADStream
from outspeed.utils.vad import VADState


class GroqLLM(Plugin):
    """
    A plugin for interacting with Groq's LLM API.

    This class handles streaming chat completions, manages chat history,
    and provides methods for running the LLM and handling interruptions.
    """

    def __init__(
        self,
        model: str = "llama-3.1-8b-instant",
        api_key: Optional[str] = None,
        base_url: Optional[str] = None,
        system_prompt: Optional[str] = None,
        stream: bool = True,
        temperature: float = 1.0,
        response_format: Optional[Dict[str, Any]] = None,
    ):
        """
        Initialize the GroqLLM plugin.

        Args:
            model (str): The name of the Groq model to use.
            api_key (Optional[str]): The API key for Groq. If None, it will be read from the environment.
            base_url (Optional[str]): The base URL for the Groq API. Not used in the current implementation.
            system_prompt (Optional[str]): The system prompt to be used in the conversation.
            stream (bool): Whether to stream the response or not.
            temperature (float): The temperature parameter for the LLM.
            response_format (Optional[Dict[str, Any]]): The desired response format.
        """
        super().__init__()
        self._model: str = model
        self._api_key: str = api_key or os.getenv("GROQ_API_KEY")
        if not self._api_key:
            raise ValueError("Groq API key is required")
        self._client: AsyncOpenAI = AsyncOpenAI(api_key=self._api_key, base_url="https://api.groq.com/openai/v1")
        self._history: List[Dict[str, str]] = []
        self.output_queue: TextStream = TextStream()
        self.chat_history_queue: TextStream = TextStream()
        self._generating: bool = False
        self._stream: bool = stream
        self._response_format: Optional[Dict[str, Any]] = response_format
        self._system_prompt: Optional[str] = system_prompt
        if self._system_prompt is not None:
            self._history.append({"role": "system", "content": self._system_prompt})
        self._temperature: float = temperature

        # These will be set later
        self.input_queue: Optional[TextStream] = None
        self.interrupt_queue: Optional[asyncio.Queue] = None
        self._task: Optional[asyncio.Task] = None
        self._interrupt_task: Optional[asyncio.Task] = None

    async def _stream_chat_completions(self) -> None:
        """
        Asynchronously stream chat completions from the Groq API.

        This method continuously reads from the input queue, sends requests to the Groq API,
        and writes the responses to the output queue.
        """
        while True:
            text_chunk: Optional[str] = await self.input_queue.get()
            if text_chunk is None:
                continue
            if isinstance(text_chunk, SessionData):
                await self.output_queue.put(text_chunk)
                continue
            self._generating = True
            tracing.register_event(tracing.Event.LLM_START)

            # Add user message to history and chat history queue
            self._history.append({"role": "user", "content": text_chunk})
            self.chat_history_queue.put_nowait(json.dumps(self._history[-1]))

            # Create chat completion request
            completion_kwargs: Dict[str, Any] = {
                "model": self._model,
                "stream": self._stream,
                "messages": self._history,
                "temperature": self._temperature,
            }
            if self._response_format:
                completion_kwargs["response_format"] = self._response_format

            chunk_stream = await self._client.chat.completions.create(**completion_kwargs)

            # Prepare for assistant's response
            self._history.append({"role": "assistant", "content": ""})
            first_chunk = True

            if self._stream:
                async for chunk in chunk_stream:
                    if first_chunk:
                        tracing.register_event(tracing.Event.LLM_TTFB)
                        first_chunk = False
                    if len(chunk.choices) == 0:
                        continue
                    elif chunk.choices[0].delta.content:
                        self._history[-1]["content"] += chunk.choices[0].delta.content
                        await self.output_queue.put(chunk.choices[0].delta.content)
            else:
                self._history[-1]["content"] = chunk_stream.choices[0].message.content
                await self.output_queue.put(chunk_stream.choices[0].message.content)
                tracing.register_event(tracing.Event.LLM_TTFB)

            tracing.register_event(tracing.Event.LLM_END)
            tracing.register_metric(tracing.Metric.LLM_TOTAL_BYTES, len(self._history[-1]["content"]))
            print("llm", self._history[-1]["content"])
            self.chat_history_queue.put_nowait(json.dumps(self._history[-1]))
            self._generating = False
            await self.output_queue.put(None)

    def run(self, input_queue: TextStream) -> Tuple[TextStream, TextStream]:
        """
        Start the LLM plugin.

        Args:
            input_queue (TextStream): The input queue to read from.

        Returns:
            Tuple[TextStream, TextStream]: The output queue and chat history queue.
        """
        self.input_queue = input_queue
        self._task = asyncio.create_task(self._stream_chat_completions())
        return self.output_queue, self.chat_history_queue

    async def close(self) -> None:
        """
        Close the LLM plugin by cancelling the main task.
        """
        if self._task:
            self._task.cancel()

    async def _interrupt(self):
        while True:
            vad_state: VADState = await self.interrupt_queue.get()
            if vad_state == VADState.SPEAKING and (not self.input_queue.empty() or not self.output_queue.empty()):
                self._task.cancel()
                try:
                    await self._task
                except asyncio.CancelledError:
                    pass
                while not self.output_queue.empty():
                    self.output_queue.get_nowait()
                while not self.input_queue.empty():
                    self.input_queue.get_nowait()
                logging.info("Done cancelling LLM")
                self._generating = False
                self._task = asyncio.create_task(self._stream_chat_completions())

    def set_interrupt_stream(self, interrupt_stream: VADStream):
        if isinstance(interrupt_stream, VADStream):
            self.interrupt_queue = interrupt_stream
        else:
            raise ValueError("Interrupt stream must be a VADStream")
        self._interrupt_task = asyncio.create_task(self._interrupt())
